@@ -493,22 +493,6 @@ field<mat> linpred_surv (const field<mat> &X, const field<vec> &betas,
   return out;
 }
 
-field<mat> create_Wlong (const field<mat> &eta, const field<uvec> &FunForms,
-                         const field<mat> &U, const field<uvec> &ind) {
-  uword n_outcomes = eta.n_elem;
-  field<mat> out(n_outcomes);
-  for (uword i = 0; i < n_outcomes; ++i) {
-    mat eta_i = eta.at(i);
-    uvec FF_i = FunForms.at(i);
-    mat U_i = U.at(i);
-    uvec ind_i = ind.at(i);
-    mat Wlong_i(eta_i.n_rows, U_i.n_cols, fill::ones);
-    Wlong_i.cols(FF_i) %= eta_i.cols(ind_i);
-    out.at(i) = U_i % Wlong_i;
-  }
-  return out;
-}
-
 field<vec> linpred_mixed (const field<mat> &X, const field<vec> &betas,
                           const field<mat> &Z, const field<mat> &b,
                           const field<uvec> &id) {
@@ -565,14 +549,62 @@ cube chol_cube (const cube &S) {
   return out;
 }
 
+void transf_eta (mat &eta, const CharacterVector &fun_nams) {
+  uword n = eta.n_cols;
+  for (uword i = 0; i < n; i++) {
+    if (fun_nams[i] == "identity") continue;
+    if (fun_nams[i] == "expit") {
+      eta.col(i) = 1.0 / (1.0 + trunc_exp(- eta.col(i)));
+    } else if (fun_nams[i] == "exp" || fun_nams[i] == "dexp") {
+      eta.col(i) = trunc_exp(eta.col(i));
+    } else if (fun_nams[i] == "dexpit") {
+      mat pp = 1.0 / (1.0 + trunc_exp(- eta.col(i)));
+      eta.col(i) = pp * (1.0 - pp);
+    } else if (fun_nams[i] == "log") {
+      eta.col(i) = trunc_log(eta.col(i));
+    } else if (fun_nams[i] == "log2") {
+      eta.col(i) = log2(eta.col(i));
+    } else if (fun_nams[i] == "log10") {
+      eta.col(i) = log10(eta.col(i));
+    } else if (fun_nams[i] == "sqrt") {
+      eta.col(i) = sqrt(eta.col(i));
+    } else if (fun_nams[i] == "square") {
+      eta.col(i) = square(eta.col(i));
+    }
+  }
+}
+
+field<mat> create_Wlong (const field<mat> &eta, const field<uvec> &FunForms,
+                         const field<mat> &U, const field<uvec> &ind,
+                         const List &Funs_FunForms) {
+  uword n_outcomes = eta.n_elem;
+  field<mat> out(n_outcomes);
+  for (uword i = 0; i < n_outcomes; ++i) {
+    CharacterVector Funs_i = Funs_FunForms[i];
+    mat eta_i = eta.at(i);
+    transf_eta(eta_i, Funs_i);
+    uvec FF_i = FunForms.at(i);
+    mat U_i = U.at(i);
+    uvec ind_i = ind.at(i);
+    mat Wlong_i(eta_i.n_rows, U_i.n_cols, fill::ones);
+    Wlong_i.cols(FF_i) %= eta_i.cols(ind_i);
+    out.at(i) = U_i % Wlong_i;
+  }
+  return out;
+}
+
 mat calculate_Wlong (const field<mat> &X, const field<mat> &Z,
                      const field<mat> &U, const mat &Wlong_bar,
+                     const mat &Wlong_sds,
                      const field<vec> &betas, const field<mat> &b,
                      const uvec &id, const field<uvec> &FunForms,
-                     const field<uvec> &FunForms_ind) {
+                     const field<uvec> &FunForms_ind,
+                     const List &Funs_FunForms) {
   field<mat> eta = linpred_surv(X, betas, Z, b, id);
-  mat Wlong = docall_cbindF(create_Wlong(eta, FunForms, U, FunForms_ind));
+  mat Wlong =
+    docall_cbindF(create_Wlong(eta, FunForms, U, FunForms_ind, Funs_FunForms));
   Wlong.each_row() -= Wlong_bar;
+  Wlong.each_row() /= Wlong_sds;
   return Wlong;
 }
 
